/*
 * libpcap-style grammar for building custom filters.
 *
 * (C) 2011 Florian Westphal <fw@strlen.de>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 */
%{
#include <errno.h>
#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>

#include <arpa/inet.h>
#include <netdb.h>

#include "parser_common.h"
#include "bsf_filter.yy.h"
#include "ct_opts.h"

extern char *yytext;
extern void parse_bsf_string(const char *bsf);

static struct nfct_filter *filter;

enum logic_prefix {
	LOG_UNDEF,
	LOG_POSITIVE,
	LOG_NEGATIVE
};

static enum logic_prefix logic_prefixes[NFCT_FILTER_MAX];
static const char *rule_str(enum nfct_filter_attr t)
{
	switch (t) {
	case NFCT_FILTER_L4PROTO: return "protocols";
	case NFCT_FILTER_SRC_IPV4: return "ipv4 source addresses";
	case NFCT_FILTER_SRC_IPV6: return "ipv6 source addresses";
	case NFCT_FILTER_DST_IPV4: return "ipv4 destination addresses";
	case NFCT_FILTER_DST_IPV6: return "ipv6 destination addresses";
	case NFCT_FILTER_L4PROTO_STATE: /* fallthrough */
	default:
		return "";
	}
}

static void verify_prefix(enum nfct_filter_attr t, enum logic_prefix prefix)
{
	switch (logic_prefixes[t]) {
	case LOG_UNDEF:
		logic_prefixes[t] = prefix;
		return;
	case LOG_POSITIVE:
		if (prefix == LOG_POSITIVE)
			return;
		break;
	case LOG_NEGATIVE:
		if (prefix == LOG_NEGATIVE)
			return;
		break;
	}
	fprintf(stderr, "Warning: filter expression: all %s will be negated\n", rule_str(t));
	sleep(2);
}

static void convert_mask6(uint32_t *dst, uint32_t mask)
{
	int i;

	for (i=0; i < 3 && mask >= 32; i++, mask -= 32)
		dst[i] = ~0;

	if (mask)
		dst[i++] = 0xffffffff << (32-mask);

	while (i < 4)
		dst[i++] = 0;
}


void yyerror(struct filter_expr_state *state, const char *str)
{
        fprintf(stderr, "filter parse error (%s): symbol %s\n", yytext, str);
	(void) state;
	exit(1);
}

bool bsf_filter_create(int fd)
{
	bool ret;

	filter = nfct_filter_create();
	if (!filter) {
		perror("nfct_create_filter");
		return false;
	}

	if (cttop_options->ct_event_filter.bsf_str)
		parse_bsf_string(cttop_options->ct_event_filter.bsf_str);

	if (nfct_filter_attach(fd, filter)) {
		perror("nfct_filter_attach");
		ret = false;
		goto out;
	}

	ret = true;
 out:
	nfct_filter_destroy(filter);
	return ret;
}

static void convert_and_free(char *addr, int af, uint32_t *result)
{
	int i, ret = inet_pton(af, addr, result);
	if (ret < 0) {
		fprintf(stderr, "cannot convert %s: %s\n",
					addr, strerror(errno));
		exit(1);
	} else if (ret == 0) {
		fprintf(stderr, "%s: not a valid %s address\n",
			addr, af == AF_INET6 ? "ipv6":"");
		exit(1);
	}
	free(addr);

	/* BSF wants host byte order */
	switch (af) {
	case AF_INET:
		*result = ntohl(*result);
		break;
	case AF_INET6:
		for (i=0; i < 4; i++, result++)
			*result = ntohl(*result);
		break;
	}
}

static void set_filter(struct filter_expr_state *state, char *addr, uint32_t mask)
{
	union {
		struct nfct_filter_ipv4 f4;
		struct nfct_filter_ipv6 f6;
	} filter_data;

	switch (state->family) {
	case AF_INET:
		convert_and_free(addr, state->family, &filter_data.f4.addr);
		filter_data.f4.mask = mask ? (0xffffffff  << (32 - mask)) : 0;
		if (state->direction == SRC)
			state->attr = NFCT_FILTER_SRC_IPV4;
		else
			state->attr = NFCT_FILTER_DST_IPV4;
		nfct_filter_add_attr(filter, state->attr, &filter_data.f4);
	break;
	case AF_INET6:
		convert_and_free(addr, state->family, filter_data.f6.addr);
		convert_mask6(filter_data.f6.mask, mask);
		if (state->direction == SRC)
			state->attr = NFCT_FILTER_SRC_IPV6;
		else
			state->attr = NFCT_FILTER_DST_IPV6;
		nfct_filter_add_attr(filter, state->attr, &filter_data.f6);
	break;
	default:
		fprintf(stderr, "address \"%s\": unknown family %d\n", addr, state->family);
		exit(1);
	}
}

%}

%parse-param	{ struct filter_expr_state *state }

%union
{
	int number;
	char *str;
}

%token <number> DIGIT
%token <str> V6ADDR
%token <str> V4ADDR
%token <str> WORD
%token L4PROTO SRC DST SLASH NOT OP_OR

%%

filters: /* empty */
        | filters filter
        ;

filter: command {
		verify_prefix(state->attr, LOG_POSITIVE);
	}
	| NOT command {
		verify_prefix(state->attr, LOG_NEGATIVE);
		if (nfct_filter_set_logic(filter, state->attr, NFCT_FILTER_LOGIC_NEGATIVE) &&
			errno != EBUSY) {
				perror("nfct_filter_set_logic");
				exit(1);
		}
	}
	;

command:
	addrcmd
	| protocol
	;

protocol:
	L4PROTO DIGIT {
		nfct_filter_add_attr_u32(filter, NFCT_FILTER_L4PROTO, $2);
		state->attr = NFCT_FILTER_L4PROTO;
	}
	| WORD {
		struct protoent *proto;

		proto = getprotobyname($1);
		if (!proto) {
			fprintf(stderr, "unknown protocol name %s\n", $1);
			exit(1);
		}
		nfct_filter_add_attr_u32(filter, NFCT_FILTER_L4PROTO, proto->p_proto);
		state->attr = NFCT_FILTER_L4PROTO;
	}
	| OP_OR protocol
	;

addrcmd:
        SRC address
	| DST address
	;

address:
	V4ADDR SLASH DIGIT {
		if ($3 <= 0 || $3 > 32) {
			fprintf(stderr, "mask %d invalid, must be within range 1-32\n", $3);
			exit(1);
		}
		state->family = AF_INET;
		set_filter(state, $1, $3);
	}
	| V4ADDR {
		state->family = AF_INET;
		set_filter(state, $1, 32);
	}
	| V6ADDR SLASH DIGIT {
		if ($3 <= 0 || $3 > 128) {
			fprintf(stderr, "mask %d invalid, must be within range 1-128\n", $3);
			exit(1);
		}
		state->family = AF_INET6;
		set_filter(state, $1, $3);
	}
	| V6ADDR {
		state->family = AF_INET6;
		set_filter(state, $1, 128);
	}
	;

