/*
 * (C) 2011 Florian Westphal <fw@strlen.de>
 * (C) 2011 Astaro GmbH & Co. KG
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>

#include "cttop.h"
#include "ct_acct.h"
#include "ct_entry.h"
#include "jhash3.h"

#include "ct_opts.h"
/**
 * @file ct_acct.c
 * keep track of how many flows and source address have
 * been associated with a particular ct_entry group.
 */

/**
 * @struct ct_acct_addr
 */
struct ct_acct_addr {
	uint8_t family; /*!< address family (AF_INET or AF_INET6) */
	uint32_t address[4]; /*!< space for single ipv4 or ipv6 address */
	uint32_t count;	/*!< how often this address has been added to set */
};


/**
 * convert_nfct_to_acct_saddr - convert struct nf_conntrack to internal representation
 * @param ct - nf_conntrack to convert
 * @param addr - converted result
 */
static void convert_nfct_to_acct_saddr(const struct nf_conntrack *ct, struct ct_acct_addr *addr)
{

	if (nfct_attr_is_set(ct, ATTR_IPV6_SRC)) {
		memcpy(addr->address, nfct_get_attr(ct, ATTR_IPV6_SRC),
					sizeof(addr->address));
		addr->family = AF_INET6;
		return;
	}
	addr->address[0] = nfct_get_attr_u32(ct, ATTR_IPV4_SRC);
	addr->family = AF_INET;
}

static guint ct_acct_hash(gconstpointer arg)
{
	const struct ct_acct_addr *a = arg;

	if (a->family == AF_INET)
		return uhash_words(a->address, 1, 0);

	return uhash_words(a->address, ARRAY_SIZE(a->address), 0);
}

/**
 * ct_acct_equal - compare two ct_acct_addr structures
 * @return true if equal
 */
static gboolean ct_acct_equal(gconstpointer a, gconstpointer b)
{
	const struct ct_acct_addr *a1 = a;
	const struct ct_acct_addr *a2 = b;

	if (a1->address[0] != a2->address[0])
		return false;
	if (a1->family != a2->family)
		return false;
	if (a1->family == AF_INET)
		return true;

	if ((a1->address[3] != a2->address[3]) ||
	    (a1->address[2] != a2->address[2]))
		return false;

	return a1->address[1] == a2->address[1];
}

/**
 * acct_del_src - remove connection from source address accounting set
 * @param ct conntrack to remove
 * @param acct accounting stat
 */
static void acct_del_src(const struct nf_conntrack *ct, struct ct_acct *acct)
{
	struct ct_acct_addr key;
	uint32_t *count;

	if (!acct->src_addrs)
		return;

	convert_nfct_to_acct_saddr(ct, &key);

	count = g_hash_table_lookup(acct->src_addrs, &key);
	if (!count) {
		fprintf(stderr, "%s: internal error: tried to remove non-existant ct\n",
									__func__);
		return;
	}

	--*count;
	if (*count == 0)
		g_hash_table_remove(acct->src_addrs, &key);
	if (acct->flows == 0)
		g_hash_table_destroy(acct->src_addrs);
}

/**
 * ct_del_conntrack - del nfct pointer from accounting set
 * @param ct conntrack pointer to remove
 * @param acct accounting stat
 */
static void acct_del_conntrack(const struct nf_conntrack *ct, struct ct_acct *acct)
{
	if (!acct->conntracks)
		return;

	g_hash_table_remove(acct->conntracks, ct);
	if (acct->flows == 0) {
		if (g_hash_table_size(acct->conntracks))
			fprintf(stderr, "%s: internal error: conntracks not empty\n",
									__func__);
		g_hash_table_destroy(acct->conntracks);
	}
}

/**
 * ct_acct_del - remove connection ct from accounting stat
 * @param ct conntrack to remove
 * @param acct accounting stat
 * @return true if ct was the last connection accounted
 */
bool ct_acct_del(const struct nf_conntrack *ct, struct ct_acct *acct)
{
	if (acct->flows)
		acct->flows--;

	acct_del_src(ct, acct);
	acct_del_conntrack(ct, acct);

	return acct->flows == 0;
}

/**
 * ct_acct_add - add connection to source address accounting set
 * @param ct conntrack to add
 * @param acct accounting stat
 */
static void acct_add_src(const struct nf_conntrack *ct, struct ct_acct *acct)
{
	struct ct_acct_addr key;
	struct ct_acct_addr *addr;
	uint32_t *count;

	if (!acct->src_addrs)
		return;

	convert_nfct_to_acct_saddr(ct, &key);
	count = g_hash_table_lookup(acct->src_addrs, &key);
	if (count) {
		++*count;
		return;
	}

	addr = malloc(sizeof(*addr));
	if (!addr) {
		perror("malloc");
		exit(1);
	}
	memcpy(addr, &key, sizeof(*addr));
	addr->count = 1;
	g_hash_table_insert(acct->src_addrs, addr, &addr->count);
}


/**
 * ct_acct_conntrack - add nfct pointer to accounting set
 * @param ct conntrack pointer to add
 * @param acct accounting stat
 */
static void acct_add_conntrack(const struct nf_conntrack *ct, struct ct_acct *acct)
{
	const struct nf_conntrack *present;

	if (!acct->conntracks)
		return;

	present = g_hash_table_lookup(acct->conntracks, ct);
	if (!present)
		g_hash_table_insert(acct->conntracks, (void*) ct, (void *) ct);
}

/**
 * ct_acct_add - add connection ct to accounting stat
 * @param ct conntrack to remove
 * @param acct accounting stat
 */
void ct_acct_add(const struct nf_conntrack *ct, struct ct_acct *acct)
{
	acct->flows++;
	acct_add_src(ct, acct);
	acct_add_conntrack(ct, acct);
}

static bool ct_acct_sync_ctmark(struct nf_conntrack *ct, struct ct_acct *acct)
{
	GHashTableIter iter;
	gpointer key, value;
	uint32_t ctmark = 0, mask;
	unsigned int i = 0;

	if (!acct->conntracks)
		return false;

	mask = cttop_options->ctmark_group_mask;

	g_hash_table_iter_init(&iter, acct->conntracks);

	while (g_hash_table_iter_next(&iter, &key, &value)) {
		uint32_t mark = nfct_get_attr_u32(key, ATTR_MARK) & mask;
		i++;
		if (i == 1) {
			ctmark = mark;
			continue;
		}
		if (ctmark != mark)
			return false;
	}

	assert(i > 0);
	nfct_set_attr_u32(ct, ATTR_MARK, ctmark);
	return true;
}

bool ct_acct_resync(struct nf_conntrack *ct, struct ct_acct *acct)
{
	return ct_acct_sync_ctmark(ct, acct);
}

/**
 * initialize acct structure and add ct as first member.
 * @param ct conntrack to add
 * @param acct accounting structure to initialize
 * @param flags additional properties that should be accounted
 * flags:
 * CTT_GRP_ATTR_SRCCOUNT: account number of source addresses.
 */
void ct_acct_init(const struct nf_conntrack *ct,
		struct ct_acct *acct, unsigned int flags)
{
	acct->conntracks = NULL;
	acct->src_addrs = NULL;
	acct->flows = 0;

	if (flags & CTT_GRP_ATTR_SRCCOUNT) {
		acct->src_addrs = g_hash_table_new_full(
				ct_acct_hash, ct_acct_equal,
				free, NULL);

		if (!acct->src_addrs) {
			perror("g_hash_table_new_full");
			exit(1);
		}
	}

	if (flags & CTT_GRP_ATTR_TRACK_CTMARK) {
		acct->conntracks = g_hash_table_new(g_direct_hash, g_direct_equal);
		if (!acct->conntracks) {
			perror("g_hash_table_new");
			exit(1);
		}
	}

	ct_acct_add(ct, acct);
}

