/*
 * (C) 2010-2011 Florian Westphal <fw@strlen.de>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>

#include <glib.h>

#include "ui.h"
#include "cttop.h"
#include "ct_acct.h"
#include "ct_entry.h"
#include "ct_opts.h"
#include "ct_events.h"
#include "ct_timer.h"
#include "jhash3.h"


/**
 * @struct list_head
 * single-linked list with tail pointer
 *
 * Used to add elements to the end without
 * walking the entire list.
 */
struct list_head {
	GSList *list;	/*<! start of list> */
	GSList *tailp;  /*<! ptr to last element in list */
};

/**
 * @struct counter_poll_stat
 * counter update polling list
 * This is a list of all known individual ct_entry structures.
 * cttop will poll the kernel periodically for byte/packet
 * counter updates for these connections.
 *
 * entries are moved from current to completed list; new connections
 * are saved on 'new' list.
 * When current becomes empty, completed becomes the new current list,
 * and new list is merged to current list.
 */
static struct {
	GSList *current; /*!< active ct_entry structures to poll for updates */
	struct list_head completed; /*!< ct_entry structures that have been polled */
	struct list_head new; /*!< new ct_entry structures that should be polled on next cycle */
} counter_poll_stat;

#define MAX_ENTRIES	100
#define AVERAGE_SECONDS	3.0

/**
 * @struct ctt_group
 * represent a flow group abstraction.
 * cttop puts flows into multiple "groups", where entries are kept in
 * aggregated state (for example, GROUP_SRC represents each known source
 * address seen, i.e.  its hash table contains a ct_entry structure for
 * each known source address).
 */
struct ctt_group {
	gboolean (*equalfn)(gconstpointer, gconstpointer);
	guint (*hashfn)(gconstpointer);
	void (*valfreefn)(void *);
	const char *name;
	GPtrArray *parray; /*!< array referencing the top ct_entry structures joined into this group by current sort order */
	GHashTable *htable; /*!< hash table containing all ct_entry structures joined into this group */
	unsigned int attr; /*!< attributes that can be retrieved from the group */
};

static uint32_t jhash_seed;

static enum sort_order sort_method = SORT_BYTES_TOTAL;
static enum sort_flags sort_flags;
static enum group_type group_active = GROUP_NONE;

enum status_bit {
	STATUS_QUERIED = (1 << 0),
	STATUS_IN_ARRAY = (1 << 1),
	STATUS_DESTROYED = (1 << 2),
	STATUS_CTMARK_CONFLICT = (1 << 3),
	STATUS_CTMARK_NEED_SYNC = (1 << 4),

	/* MAX: 1 << 15 */
};

enum acct_dir {
	DIR_IN,
	DIR_OUT
};

struct byte_stats {
	uint64_t bytes[2];
};

struct ct_entry {
	struct cttop_stats stats;
	struct byte_stats avg;
	struct byte_stats avg_tmp;
	unsigned int avg_samples;
	uint16_t tick_count;
	uint16_t flags;
	struct ct_acct acct;
	uint64_t modified_tick;
	struct nf_conntrack *nfct;
};

static void list_append(struct list_head *h, GSList *append)
{
	if (h->tailp)
		h->tailp->next = append;
	else
		h->list = append;

	while (append->next)
		append = append->next;
	h->tailp = append;
}

static bool list_append_new(struct list_head *h, void *data)
{
	GSList *list = g_slist_alloc();
	if (!list) {
		perror("g_slist_append");
		return false;
	}
	list->data = data;
	list_append(h, list);
	return true;
}

static void list_null(struct list_head *h)
{
	h->list = NULL;
	h->tailp = NULL;
}

static bool ct_entry_has_flag(const struct ct_entry *e, enum status_bit bit)
{
	return (e->flags & bit) != 0;
}
static void ct_entry_set_flag(struct ct_entry *e, enum status_bit bit)
{
	e->flags |= bit;
}
static void ct_entry_unset_flag(struct ct_entry *e, enum status_bit bit)
{
	e->flags &= ~bit;
}

static gint ctentry_cmp_stats_u64(const uint64_t *a, const uint64_t *b)
{
	int ret;

	if (*a < *b)
		ret = 1;
	else if (*a > *b)
		ret = -1;
	else
		return 0;
	if (sort_flags == SORT_FLAG_REVERSE)
		return -ret;
	return ret;
}

static gint ctentry_cmp_stats_inoutb(const struct ct_entry *c1, const struct ct_entry *c2)
{
	uint64_t t1 = c1->stats.in.bytes + c1->stats.out.bytes;
	uint64_t t2 = c2->stats.in.bytes + c2->stats.out.bytes;
	return ctentry_cmp_stats_u64(&t1, &t2);
}

static gint ctentry_cmp_stats_inb(const struct ct_entry *c1, const struct ct_entry *c2)
{
	return ctentry_cmp_stats_u64(&c1->stats.in.bytes, &c2->stats.in.bytes);
}

static gint ctentry_cmp_stats_outb(const struct ct_entry *c1, const struct ct_entry *c2)
{
	return ctentry_cmp_stats_u64(&c1->stats.out.bytes, &c2->stats.out.bytes);
}

static gint ctentry_cmp_avg_inoutb(const struct ct_entry *c1, const struct ct_entry *c2)
{
	uint64_t t1 = c1->avg.bytes[DIR_IN] + c1->avg.bytes[DIR_OUT];
	uint64_t t2 = c2->avg.bytes[DIR_IN] + c2->avg.bytes[DIR_OUT];
	return ctentry_cmp_stats_u64(&t1, &t2);
}

static gint ctentry_cmp_avg_inb(const struct ct_entry *c1, const struct ct_entry *c2)
{
	return ctentry_cmp_stats_u64(&c1->avg.bytes[DIR_IN], &c2->avg.bytes[DIR_IN]);
}

static gint ctentry_cmp_avg_outb(const struct ct_entry *c1, const struct ct_entry *c2)
{
	return ctentry_cmp_stats_u64(&c1->avg.bytes[DIR_OUT], &c2->avg.bytes[DIR_OUT]);
}

static const struct {
	gint(*sort_func)(const struct ct_entry *a, const struct ct_entry *b);
	const char *name;
} sort_methods[] = {
	[SORT_BYTES_TOTAL] = {
		.sort_func = ctentry_cmp_stats_inoutb,
		.name = "bytes total",
	},
	[SORT_BYTES_IN] = {
		.sort_func = ctentry_cmp_stats_inb,
		.name = "bytes in",
	},
	[SORT_BYTES_OUT] = {
		.sort_func = ctentry_cmp_stats_outb,
		.name = "bytes out",
	},
	[SORT_BYTES_AVG_TOTAL] = {
		.sort_func = ctentry_cmp_avg_inoutb,
		.name = "bytes average",
	},
	[SORT_BYTES_AVG_IN] = {
		.sort_func = ctentry_cmp_avg_inb,
		.name = "bytes in average",
	},
	[SORT_BYTES_AVG_OUT] = {
		.sort_func = ctentry_cmp_avg_outb,
		.name = "bytes out average",
	},
};

static uint32_t do_uhash(const uint32_t *keys, size_t num_keys)
{
	return uhash_words(keys, num_keys, jhash_seed);
}

static uint32_t nfct_get_last_ipv6_word(const struct nf_conntrack *ct,
				const enum nf_conntrack_attr type)
{
	const uint32_t *tmp = nfct_get_attr(ct, type);
	if (!tmp) {
		perror("nfct_get_attr ATTR_IPV6");
		return 0;
	}
	return tmp[3];
}

static guint nfct_hash_dport(gconstpointer arg)
{
	const struct nf_conntrack *ct = arg;
	uint32_t scratch[3];
	unsigned int family;

	family = nfct_get_attr_u8(ct, ATTR_L3PROTO);
	scratch[0] = (family << 8) | nfct_get_attr_u8(ct, ATTR_L4PROTO);
	scratch[0] <<= 16;

	scratch[0] += nfct_get_attr_u16(ct, ATTR_PORT_DST);

	if (family == AF_INET) {
		scratch[1] = nfct_get_attr_u32(ct, ATTR_IPV4_SRC);
		scratch[2] = nfct_get_attr_u32(ct, ATTR_IPV4_DST);
	} else {
		scratch[1] = nfct_get_last_ipv6_word(ct, ATTR_IPV6_SRC);
		scratch[2] = nfct_get_last_ipv6_word(ct, ATTR_IPV6_DST);
	}

	return do_uhash(scratch, ARRAY_SIZE(scratch));
}

static guint nfct_hash_dst(gconstpointer arg)
{
	const struct nf_conntrack *ct = arg;
	unsigned int family = nfct_get_attr_u8(ct, ATTR_L3PROTO);
	const uint32_t *addr;

	if (family == AF_INET) {
		addr = nfct_get_attr(ct, ATTR_IPV4_DST);
		return do_uhash(addr, 1);
	}
	addr = nfct_get_attr(ct, ATTR_IPV6_DST);
	assert(addr);
	return do_uhash(addr, 4);
}
static guint nfct_hash_src(gconstpointer arg)
{
	const struct nf_conntrack *ct = arg;
	unsigned int family = nfct_get_attr_u8(ct, ATTR_L3PROTO);
	const uint32_t *addr;

	if (family == AF_INET) {
		addr = nfct_get_attr(ct, ATTR_IPV4_SRC);
		return do_uhash(addr, 1);
	}
	addr = nfct_get_attr(ct, ATTR_IPV6_SRC);
	assert(addr);
	return do_uhash(addr, 4);
}

static guint nfct_hash(gconstpointer arg)
{
	const struct nf_conntrack *ct = arg;
	uint32_t scratch[4];
	unsigned int family;

	family = nfct_get_attr_u8(ct, ATTR_L3PROTO);
	scratch[0] = nfct_get_attr_u16(ct, ATTR_PORT_DST) << 16 |
			nfct_get_attr_u16(ct, ATTR_PORT_SRC);
	scratch[1] = (family << 8) | nfct_get_attr_u8(ct, ATTR_L4PROTO);

	if (family == AF_INET) {
		scratch[2] = nfct_get_attr_u32(ct, ATTR_IPV4_SRC);
		scratch[3] = nfct_get_attr_u32(ct, ATTR_IPV4_DST);
	} else {
		scratch[2] = nfct_get_last_ipv6_word(ct, ATTR_IPV6_SRC);
		scratch[3] = nfct_get_last_ipv6_word(ct, ATTR_IPV6_DST);
	}

	return do_uhash(scratch, ARRAY_SIZE(scratch));
}

static guint nfct_hash_ctmark(gconstpointer arg)
{
	uint32_t mask = cttop_options->ctmark_group_mask;
	uint32_t val = nfct_get_attr_u32(arg, ATTR_MARK) & mask;
	return do_uhash(&val, 1);
}

static gint ctentry_nfct_cmp(gconstpointer a, gconstpointer b)
{
	const struct nf_conntrack *ct1 = a;
	const struct nf_conntrack *ct2 = b;

	return nfct_cmp(ct1, ct2, NFCT_CMP_ORIG);
}

static gboolean ctentry_dst_equal(gconstpointer a, gconstpointer b)
{
	const struct nf_conntrack *ct1 = a;
	const struct nf_conntrack *ct2 = b;
	unsigned int family;
	const uint32_t *a1, *a2;

	family = nfct_get_attr_u8(ct1, ATTR_L3PROTO);
	if (family != nfct_get_attr_u8(ct2, ATTR_L3PROTO))
		return false;

	if (family == AF_INET) {
		return nfct_get_attr_u32(ct1, ATTR_IPV4_DST) ==
			nfct_get_attr_u32(ct2, ATTR_IPV4_DST);
	}
	a1 = nfct_get_attr(ct1, ATTR_IPV6_DST);
	a2 = nfct_get_attr(ct2, ATTR_IPV6_DST);
	return (memcmp(a1, a2, sizeof(uint32_t) * 4)) == 0;
}

static gboolean ctentry_src_equal(gconstpointer a, gconstpointer b)
{
	const struct nf_conntrack *ct1 = a;
	const struct nf_conntrack *ct2 = b;
	unsigned int family;
	const uint32_t *a1, *a2;

	family = nfct_get_attr_u8(ct1, ATTR_L3PROTO);
	if (family != nfct_get_attr_u8(ct2, ATTR_L3PROTO))
		return false;

	if (family == AF_INET) {
		return nfct_get_attr_u32(ct1, ATTR_IPV4_SRC) ==
			nfct_get_attr_u32(ct2, ATTR_IPV4_SRC);
	}
	a1 = nfct_get_attr(ct1, ATTR_IPV6_SRC);
	a2 = nfct_get_attr(ct2, ATTR_IPV6_SRC);
	return (memcmp(a1, a2, sizeof(uint32_t) * 4)) == 0;
}

static gboolean ctentry_dport_equal(gconstpointer a, gconstpointer b)
{
	const struct nf_conntrack *ct1 = a;
	const struct nf_conntrack *ct2 = b;

	if (nfct_get_attr_u16(ct1, ATTR_PORT_DST) !=
	    nfct_get_attr_u16(ct2, ATTR_PORT_DST))
		return false;

	if (nfct_get_attr_u8(ct1, ATTR_L4PROTO) !=
	    nfct_get_attr_u8(ct2, ATTR_L4PROTO))
		return false;

	return ctentry_src_equal(a, b) && ctentry_dst_equal(a, b);
}

static gboolean ctentry_ctmark_equal(gconstpointer a, gconstpointer b)
{
	const struct nf_conntrack *ct1 = a;
	const struct nf_conntrack *ct2 = b;
	uint32_t mask = cttop_options->ctmark_group_mask;

	return (nfct_get_attr_u32(ct1, ATTR_MARK) & mask) ==
			(nfct_get_attr_u32(ct2, ATTR_MARK) & mask);
}

static void ct_entry_free(void *p)
{
	struct ct_entry *e = p;
	nfct_destroy(e->nfct);
	free(e);
}

static struct ctt_group ctt_groups[] = {
	[GROUP_NONE] = {
		.attr = CTT_GRP_ATTR_SRC | CTT_GRP_ATTR_DST |
				CTT_GRP_ATTR_PROTO | CTT_GRP_ATTR_PORTS |
				CTT_GRP_ATTR_CTMARK | CTT_GRP_ATTR_CTSECMARK,
		.equalfn = ctentry_nfct_cmp,
		.hashfn = nfct_hash,
		.name = "all flows",
	},
	[GROUP_SRC] = {
		.attr = CTT_GRP_ATTR_SRC | CTT_GRP_ATTR_FLOWCOUNT,
		.equalfn = ctentry_src_equal,
		.hashfn = nfct_hash_src,
		.valfreefn = ct_entry_free,
		.name = "source address",
	},
	[GROUP_DST] = {
		.attr = CTT_GRP_ATTR_DST | CTT_GRP_ATTR_FLOWCOUNT,
		.equalfn = ctentry_dst_equal,
		.hashfn = nfct_hash_dst,
		.valfreefn = ct_entry_free,
		.name = "destination address",
	},
	[GROUP_DPORT] = {
		.attr = CTT_GRP_ATTR_SRC | CTT_GRP_ATTR_DST |
			CTT_GRP_ATTR_DPORT | CTT_GRP_ATTR_PROTO |
			CTT_GRP_ATTR_FLOWCOUNT | CTT_GRP_ATTR_CTMARK | CTT_GRP_ATTR_TRACK_CTMARK,
		.equalfn = ctentry_dport_equal,
		.hashfn = nfct_hash_dport,
		.valfreefn = ct_entry_free,
		.name = "source, destination, destination port",
	},
	[GROUP_CTMARK] = {
		.attr = CTT_GRP_ATTR_CTMARK | CTT_GRP_ATTR_FLOWCOUNT |
			CTT_GRP_ATTR_SRCCOUNT,
		.equalfn = ctentry_ctmark_equal,
		.hashfn = nfct_hash_ctmark,
		.valfreefn = ct_entry_free,
		.name = "conntrack mark",
	},
};

static gint ctentry_sort_function(gconstpointer a, gconstpointer b)
{
	const struct ct_entry * const *p1 = a;
	const struct ct_entry * const *p2 = b;
	const struct ct_entry *c1 = *p1;
	const struct ct_entry *c2 = *p2;

	return sort_methods[sort_method].sort_func(c1, c2);
}

static struct ct_entry *ct_entry_last_in_array(struct ctt_group *grp)
{
	struct ct_entry *e = NULL;
	int lastidx = grp->parray->len - 1;

	if (lastidx >= 0)
		e = g_ptr_array_index(grp->parray, lastidx);

	return e;
}

static void add_to_array_cond(struct ctt_group *grp, struct ct_entry *e)
{
	struct ct_entry *last;
	GPtrArray *array = grp->parray;

	if (array->len > MAX_ENTRIES) {
		last = ct_entry_last_in_array(grp);

		if (ctentry_sort_function(&e, &last) < 0)
			return;
	}

	if (!ct_entry_has_flag(e, STATUS_IN_ARRAY)) {
		ct_entry_set_flag(e, STATUS_IN_ARRAY);
		g_ptr_array_add(array, e);
	}

	assert(!ct_entry_has_flag(e, STATUS_DESTROYED));

	g_ptr_array_sort(array, ctentry_sort_function);

	if (array->len > MAX_ENTRIES) {
		last = ct_entry_last_in_array(grp);
		ct_entry_unset_flag(last, STATUS_IN_ARRAY);
		g_ptr_array_remove_index_fast(array, array->len - 1);
	}
}

/**
 * ct_entry_poll_counters - start or resume counter poll cycle
 * update connection tracking counters of known flows.
 *
 * To avoid blocking for too long, this hook returns either
 * if all counters have been updated, or if a certain limit has
 * been reached.
 * @return remaining budget (0 means budget expired).
 * @see ct_entry_poll_counters_resume
 */
unsigned int ct_entry_poll_counters(void)
{
	GSList *cur;
	unsigned int budget = 100;
 again:
	while ((cur = counter_poll_stat.current)) {
		struct ct_entry *entry = cur->data;

		if (ct_entry_has_flag(entry, STATUS_DESTROYED)) {
			counter_poll_stat.current = g_slist_next(counter_poll_stat.current);
			ct_entry_free(entry);
			cur->next = NULL;
			g_slist_free(cur);
			continue;
		}

		if (ct_entry_has_flag(entry, STATUS_QUERIED)) {
			ct_entry_del(entry->nfct);
			goto skip;
		}

		/* must set this before query is sent */
		ct_entry_set_flag(entry, STATUS_QUERIED);

		if (!ctevent_query(entry->nfct)) {
			ct_entry_unset_flag(entry, STATUS_QUERIED);
			break;
		}
 skip:
		counter_poll_stat.current = g_slist_next(counter_poll_stat.current);
		cur->next = NULL;
		list_append(&counter_poll_stat.completed, cur);

		if (--budget == 0)
			break;
	}

	if (counter_poll_stat.current == NULL) {
		bool restart = false;
		GSList *new = counter_poll_stat.new.list;
		if (new) {
			counter_poll_stat.new.tailp->next = counter_poll_stat.completed.list;
			list_null(&counter_poll_stat.new);

			/*
			 * unlikely: nothing processed, but new arrivals.
			 * Process them right away, this avoids delaying their
			 * counter update for a whole tick interval.
			 */
			if (counter_poll_stat.completed.list == NULL)
				restart = true;
			counter_poll_stat.completed.list = new;
		} else {
			new = counter_poll_stat.completed.list;
		}

		counter_poll_stat.current = new;
		list_null(&counter_poll_stat.completed);
		if (restart)
			goto again;
	}

	return budget;
}

/**
 * ct_entry_poll_counters_resume - continue an interrupted poll cycle
 *
 * If all entries have been polled in this cycle this function does nothing.
 * @returns the remaining poll budget.
 */
unsigned int ct_entry_poll_counters_resume(void)
{
	unsigned int budget = ~0;

	if (counter_poll_stat.current && counter_poll_stat.completed.list)
		budget = ct_entry_poll_counters();

	return budget;
}

static struct ct_entry *ct_entry_alloc(struct nf_conntrack *ct)
{
	struct ct_entry *e;
	e = malloc(sizeof(*e));
	if (!e) {
		perror("malloc");
		return NULL;
	}

	e->nfct = nfct_clone(ct);
	if (!e->nfct) {
		perror("nfct_clone");
		free(e);
		return NULL;
	}

	memset(&e->stats, 0, sizeof(e->stats));
	memset(&e->avg, 0, sizeof(e->avg));
	memset(&e->avg_tmp, 0, sizeof(e->avg_tmp));
	e->tick_count = 0;
	e->avg_samples = 0;

	e->flags = 0;
	e->modified_tick = cttimer_get_ticks();
	return e;
}

static void group_add(int idx, struct nf_conntrack *ct)
{
	struct ct_entry *entry;
	struct ctt_group *grp = &ctt_groups[idx];

	entry = g_hash_table_lookup(grp->htable, ct);
	if (entry) {
		ct_acct_add(ct, &entry->acct);
		return;
	}

	entry = ct_entry_alloc(ct);
	if (!entry)
		return;
	g_hash_table_insert(grp->htable, entry->nfct, entry);
	ct_acct_init(ct, &entry->acct, grp->attr);
	add_to_array_cond(grp, entry);
}

static struct ct_entry *ct_entry_new2(struct nf_conntrack *ct)
{
	struct ct_entry *e;
	size_t i;

	e = ct_entry_alloc(ct);
	if (!e)
		return NULL;

	if (!list_append_new(&counter_poll_stat.new, e)) {
		free(e);
		return NULL;
	}

	ct_acct_init(ct, &e->acct, 0);

	g_hash_table_insert(ctt_groups[GROUP_NONE].htable, e->nfct, e);
	add_to_array_cond(&ctt_groups[GROUP_NONE], e);

	for (i=GROUP_NONE + 1; i < ARRAY_SIZE(ctt_groups); i++)
		group_add(i, e->nfct);
	return e;
}


/**
 * ct_entry_new - add a new connection tracking entry
 * @param ct connection tracking entry to add 
 */
void ct_entry_new(struct nf_conntrack *ct)
{
	struct ct_entry *e = g_hash_table_lookup(
			ctt_groups[GROUP_NONE].htable, ct);
	if (!e)
		ct_entry_new2(ct);
}

static struct cttop_stats ct_stats_delta(const struct cttop_stats *old, const struct cttop_stats *new)
{
	struct cttop_stats delta;
	static int warn_once;

	if (new->in.packets < old->in.packets ||
	    new->out.packets < old->out.packets) {
		if (!warn_once)
			fprintf(stderr, "packet counter shrunk (in %" PRIu64 " -> %" PRIu64 ","
							" out %" PRIu64 " -> %" PRIu64 ")\n",
				new->in.packets , old->in.packets, new->out.packets, old->out.packets);
		warn_once = 1;
		memset(&delta, 0, sizeof(delta));
		return delta;
	}

	delta.in.packets = new->in.packets - old->in.packets;
	delta.in.bytes = new->in.bytes - old->in.bytes;
	delta.out.packets = new->out.packets - old->out.packets;
	delta.out.bytes = new->out.bytes - old->out.bytes;

	return delta;
}

static void ct_counter_add(struct cttop_counter *counter, const struct cttop_counter *new)
{
	counter->packets += new->packets;
	counter->bytes += new->bytes;
}

static void ct_stats_add(struct cttop_stats *stats, const struct cttop_stats *new)
{
	ct_counter_add(&stats->in, &new->in);
	ct_counter_add(&stats->out, &new->out);
}

static uint64_t value_to_sec(uint64_t value, double ticks_per_sec)
{
	double bytes = value;
	bytes *= ticks_per_sec;
	return (uint64_t)  bytes;
}

static void update_average(struct ct_entry *e, const struct cttop_stats *delta)
{
	uint64_t now = cttimer_get_ticks();
	uint64_t tick_average;
	double ticks_per_sec = cttimer_get_ticks_per_sec();
	unsigned int avg_samples;

	if (now == 0)
		return;

	tick_average = AVERAGE_SECONDS * ticks_per_sec;
	if (tick_average == 0)
		tick_average = 1;

	e->tick_count += now - e->modified_tick;
	e->avg_samples++;

	e->avg_tmp.bytes[DIR_IN]  += delta->in.bytes;
	e->avg_tmp.bytes[DIR_OUT] += delta->out.bytes;

	if (!e->tick_count)
		return;

	e->modified_tick = now;

	avg_samples = ct_acct_get_flow_count(&e->acct) * tick_average;
	if (e->avg_samples < avg_samples)
		return;

	e->avg.bytes[DIR_IN] = value_to_sec(e->avg_tmp.bytes[DIR_IN], ticks_per_sec);
	e->avg.bytes[DIR_OUT] = value_to_sec(e->avg_tmp.bytes[DIR_OUT], ticks_per_sec);

	e->avg.bytes[DIR_IN] /= e->tick_count;
	e->avg.bytes[DIR_OUT] /= e->tick_count;

	memset(&e->avg_tmp, 0, sizeof(e->avg_tmp));
	e->tick_count = 0;
	e->avg_samples = 0;
}

static void group_update(int idx, const struct nf_conntrack *ct,
			const struct cttop_stats *delta)
{
	struct ctt_group *grp = &ctt_groups[idx];
	struct ct_entry *entry = g_hash_table_lookup(grp->htable, ct);
	if (entry) {
		ct_stats_add(&entry->stats, delta);
		update_average(entry, delta);
		add_to_array_cond(grp, entry);
	}
}

static void ct_entry_update_stats(struct ct_entry *e, const struct cttop_stats *s)
{
	struct cttop_stats delta = ct_stats_delta(&e->stats, s);
	size_t i;

	group_update(GROUP_NONE, e->nfct, &delta);

	for (i=GROUP_NONE+1;i < ARRAY_SIZE(ctt_groups); i++)
		group_update(i, e->nfct, &delta);
	e->stats = *s;
}

static void group_del(int grp, const struct nf_conntrack *ct);
static void move_mark_group(struct nf_conntrack *ct, struct ct_entry *e)
{
	uint32_t new, old, mask;
	unsigned int i;

	if (!nfct_attr_is_set(ct, ATTR_MARK))
		return;

	mask = cttop_options->ctmark_group_mask;
	old = nfct_get_attr_u32(e->nfct, ATTR_MARK) & mask;
	new = nfct_get_attr_u32(ct, ATTR_MARK) & mask;

	if (old == new)
		return;

	group_del(GROUP_CTMARK, e->nfct);
	group_add(GROUP_CTMARK, ct);
	group_update(GROUP_CTMARK, ct, &e->stats);

	if (new == 0)
		return;

	for (i=GROUP_NONE + 1; i < GROUP_MAX; i++) {

		if (i == GROUP_CTMARK)
			continue;

		e = g_hash_table_lookup(ctt_groups[i].htable, ct);
		/*
		 * its possible ct will not be member of a group, e.g.
		 * icmp will not be in _DPORT
		 */
		if (!e)
			continue;

		if (ct_entry_get_flows(e) == 1) {
			nfct_copy_attr(e->nfct, ct, ATTR_MARK);
			continue;
		}

		if (ct_entry_has_flag(e, STATUS_CTMARK_CONFLICT)) {
			ct_entry_set_flag(e, STATUS_CTMARK_NEED_SYNC);
			continue;
		}

		old = nfct_get_attr_u32(e->nfct, ATTR_MARK) & mask;
		if (old == 0)
			nfct_copy_attr(e->nfct, ct, ATTR_MARK);
		else if (old != new)
			ct_entry_set_flag(e, STATUS_CTMARK_CONFLICT);
	}
}

/**
 * ct_entry_update - update existing connection tracking entry
 * @param ct connection tracking entry to update
 * @param s current statistic counters
 *
 * if ct does not exist yet it is added automatically.
 */
void ct_entry_update(struct nf_conntrack *ct, const struct cttop_stats *s)
{
	struct ct_entry *e = g_hash_table_lookup(
			ctt_groups[GROUP_NONE].htable, ct);
	if (!e) {
		e = ct_entry_new2(ct);
		if (!e)
			return;
	} else {
		move_mark_group(ct, e);
		nfct_copy(e->nfct, ct, NFCT_CP_META);
	}

	if (s->in.packets || s->out.packets)
		ct_entry_update_stats(e, s);

	ct_entry_unset_flag(e, STATUS_QUERIED);
}

static void group_array_del(struct ctt_group *grp, const struct ct_entry *e)
{
	guint i, len = grp->parray->len;
	struct ct_entry *search;

	for (i=0; i < len; i++) {
		search = g_ptr_array_index(grp->parray, i);

		if (search == e) {
			g_ptr_array_remove_index_fast(grp->parray, i);
			break;
		}
	}
}

static void group_del(int groupidx, const struct nf_conntrack *ct)
{
	struct ct_entry *e;
	struct ctt_group *grp;

	grp = &ctt_groups[groupidx];

	e = g_hash_table_lookup(grp->htable, ct);
	if (!e)
		return;

	if (ct_acct_del(ct, &e->acct)) {
		group_array_del(grp, e);
		g_hash_table_remove(grp->htable, ct);
		return;
	}

	if (groupidx != GROUP_NONE && ct_entry_has_flag(e, STATUS_CTMARK_CONFLICT) &&
	    ct_entry_get_flows(e) == 1) {
		ct_entry_unset_flag(e, STATUS_CTMARK_CONFLICT);
		nfct_copy_attr(e->nfct, ct, ATTR_MARK);
	}
}

/**
 * ct_entry_del - delete ct entry
 * @param ct connection tracking entry to remove
 */
void ct_entry_del(struct nf_conntrack *ct)
{
	size_t i;
	struct ct_entry *e;

	e = g_hash_table_lookup(ctt_groups[GROUP_NONE].htable, ct);
	if (!e)
		return;

	assert(!ct_entry_has_flag(e, STATUS_DESTROYED));

	for (i=GROUP_NONE + 1; i < ARRAY_SIZE(ctt_groups); i++)
		group_del(i, e->nfct);
	/* e must be deleted last */
	group_del(GROUP_NONE, e->nfct);

	ct_entry_set_flag(e, STATUS_DESTROYED);
}

static void rebuild_array_cb(gpointer key, gpointer value, gpointer user_data)
{
	struct ct_entry *e = value;
	struct ctt_group *grp = user_data;

	ct_entry_unset_flag(e, STATUS_IN_ARRAY);
	add_to_array_cond(grp, e);
	(void) key;
}

static void group_rebuild_array(struct ctt_group *grp)
{
	g_ptr_array_set_size(grp->parray, 0);

	g_hash_table_foreach(grp->htable,
				rebuild_array_cb, grp);
}

static void group_rebuild_array_cond(struct ctt_group *grp)
{
	g_ptr_array_sort(grp->parray, ctentry_sort_function);
}

/**
 * ct_entry_get_nfct
 * @param e ct_entry
 * @return nf_conntrack assigned to the ct_entry
 */
const struct nf_conntrack *ct_entry_get_nfct(const struct ct_entry *e)
{
	return e->nfct;
}

/**
 * ct_entry_get_index
 * @param idx index
 * @return ct_entry at position or NULL
 */
struct ct_entry *ct_entry_get_index(unsigned int idx)
{
	GPtrArray *array;
	if (idx == 0)
		group_rebuild_array_cond(&ctt_groups[group_active]);

	array = ctt_groups[group_active].parray;

	if (idx >= array->len)
		return NULL;
	return g_ptr_array_index(array, idx);
}

/**
 * @param e ct_entry
 * @return statistic counters of the entry
 */
struct cttop_stats* ct_entry_stats(struct ct_entry *e)
{
	return &e->stats;
}

/**
 * @param e ct_entry
 * @return average input byte counter
 */
uint64_t ct_entry_avg_in(const struct ct_entry *e)
{
	return e->avg.bytes[DIR_IN];
}

/**
 * @param e ct_entry
 * @return average output byte counter
 */
uint64_t ct_entry_avg_out(const struct ct_entry *e)
{
	return e->avg.bytes[DIR_OUT];
}

/**
 * @param flags sorting flags to change
 * known flags: SORT_FLAG_REVERSE
 */
void ct_entry_toggle_sort_flags(unsigned int flags)
{
	bool changed = false;

	if (flags & SORT_FLAG_REVERSE && !(sort_flags & SORT_FLAG_REVERSE)) {
		changed = true;
		sort_flags |= SORT_FLAG_REVERSE;
	} else if (sort_flags & SORT_FLAG_REVERSE) {
		changed = true;
		sort_flags &= ~SORT_FLAG_REVERSE;
	}
	if (changed)
		group_rebuild_array(&ctt_groups[group_active]);
}

/**
 * @param type group mode view to switch to 
 */
void ct_entry_toggle_group(enum group_type type)
{
	if (type >= GROUP_MAX)
		return;

	if (type != group_active) {
		group_active = type;
	} else {
		group_active = GROUP_NONE;
	}

	if (ctt_groups[group_active].parray == NULL)
	/* can be called via option parsing before any entries exist */
		return;
	group_rebuild_array(&ctt_groups[group_active]);
}

const char *ct_entry_group_get_name(enum group_type type)
{
	assert(type < GROUP_MAX);
	return ctt_groups[type].name;
}

/**
 * ct_entry_cycle_group - switch to next group view
 * @return group mode that is now in effect
 */
enum group_type ct_entry_cycle_group(void)
{
	enum group_type next = group_active;

	next++;
	if (next == GROUP_MAX)
		next = GROUP_NONE;
	group_active = next;
	return next;
}

/**
 * @param order sorting method that should be used to display ct entries
 */
void ct_entry_set_sort_order(enum sort_order order)
{
	if (order < __MAX_SORT_ORDERS && order != sort_method) {
		sort_method = order;
		group_rebuild_array(&ctt_groups[group_active]);
	}
}

/**
 * ct_entry_cycle_sort -
 * @return sort_order that is now in effect
 */
enum sort_order ct_entry_cycle_sort(void)
{
	enum sort_order next = sort_method;

	next++;
	if (next == __MAX_SORT_ORDERS)
		next = 0;
	ct_entry_set_sort_order(next);
	return next;
}

/**
 * ct_entry_sort_get_name
 * @return pointer to readable name of the sort method
 */
const char *ct_entry_sort_get_name(enum sort_order s)
{
	assert(s < __MAX_SORT_ORDERS);
	return sort_methods[s].name;
}

static void list_free(GSList *list)
{
	GSList *tmp = list;

	for (; list ; list = g_slist_next(list))
		ct_entry_free(list->data);
	g_slist_free(tmp);
}

static void ct_entry_exit(void)
{
	size_t i;

	for (i=0; i < ARRAY_SIZE(ctt_groups); i++) {
		g_ptr_array_free(ctt_groups[i].parray, TRUE);
		g_hash_table_destroy(ctt_groups[i].htable);
	}

	list_free(counter_poll_stat.current);
	list_free(counter_poll_stat.new.list);
	list_free(counter_poll_stat.completed.list);
}

bool ct_entry_group_has_attr(enum ctt_group_attr attr)
{
	return (ctt_groups[group_active].attr & attr) == attr;
}

unsigned int ct_entry_get_flows(const struct ct_entry *e)
{
	return ct_acct_get_flow_count(&e->acct);
}

unsigned int ct_entry_get_srccount(const struct ct_entry *e)
{
	return ct_acct_get_saddr_count(&e->acct);
}

bool ct_entry_get_mark(struct ct_entry *e, uint32_t *mark)
{
	if (ct_entry_has_flag(e, STATUS_CTMARK_NEED_SYNC)) {
		if (ct_acct_resync(e->nfct, &e->acct))
			ct_entry_unset_flag(e, STATUS_CTMARK_CONFLICT);

		ct_entry_unset_flag(e, STATUS_CTMARK_NEED_SYNC);
	}
	*mark = nfct_get_attr_u32(e->nfct, ATTR_MARK);
	return !ct_entry_has_flag(e, STATUS_CTMARK_CONFLICT);
}

struct del_cb_arg {
	struct nf_conntrack *ct;
	struct nfct_handle *handle;
	unsigned int counter;
};

static void del_cb(gpointer key, gpointer value, gpointer user_data)
{
	struct del_cb_arg *args = user_data;
	struct nf_conntrack *ct = key;
	(void) value;

	if (nfct_cmp(args->ct, ct, NFCT_CMP_ALL | NFCT_CMP_MASK) &&
	    nfct_query(args->handle, NFCT_Q_DESTROY, ct) == 0)
		args->counter++;
}

unsigned int ct_entry_kill(const struct ct_entry *e)
{
	struct nf_conntrack *ct;
	struct nfct_handle *h;
	int family;
	struct del_cb_arg arg;

	h = nfct_open(CONNTRACK, 0);
        if (!h)
		return 0;

	if (group_active == GROUP_NONE) {
		if (nfct_query(h, NFCT_Q_DESTROY, e->nfct) == 0)
			arg.counter = 1;
		goto out;
	}

	ct = nfct_new();
	if (!ct)
		goto out;

	family = AF_UNSPEC;
	arg.counter = 0;
	arg.handle = h;
	arg.ct = ct;

	if (ct_entry_group_has_attr(CTT_GRP_ATTR_SRC)) {
		nfct_copy_attr(ct, e->nfct, ATTR_IPV4_SRC);
		nfct_copy_attr(ct, e->nfct, ATTR_IPV6_SRC);
		family = nfct_get_attr_u8(e->nfct, ATTR_L3PROTO);
	}

	if (ct_entry_group_has_attr(CTT_GRP_ATTR_DST)) {
		nfct_copy_attr(ct, e->nfct, ATTR_IPV4_DST);
		nfct_copy_attr(ct, e->nfct, ATTR_IPV6_DST);
		family = nfct_get_attr_u8(e->nfct, ATTR_L3PROTO);
	}

	if (family != AF_UNSPEC)
		nfct_copy_attr(ct, e->nfct, ATTR_L3PROTO);

	if (ct_entry_group_has_attr(CTT_GRP_ATTR_PROTO))
		nfct_copy_attr(ct, e->nfct, ATTR_L4PROTO);

	if (ct_entry_group_has_attr(CTT_GRP_ATTR_SPORT))
		nfct_copy_attr(ct, e->nfct, ATTR_PORT_SRC);
	if (ct_entry_group_has_attr(CTT_GRP_ATTR_DPORT))
		nfct_copy_attr(ct, e->nfct, ATTR_PORT_DST);

	if (ct_entry_group_has_attr(CTT_GRP_ATTR_CTMARK))
		nfct_copy_attr(ct, e->nfct, ATTR_MARK);

	g_hash_table_foreach(ctt_groups[GROUP_NONE].htable,
					del_cb, &arg);
	nfct_destroy(ct);
 out:
	nfct_close(h);
	return arg.counter;
}

void ct_entry_init(void)
{
	size_t i;

	jhash_seed = random();

	for (i=0; i < ARRAY_SIZE(ctt_groups); i++) {
		ctt_groups[i].htable = g_hash_table_new_full(
			ctt_groups[i].hashfn, ctt_groups[i].equalfn,
			NULL, ctt_groups[i].valfreefn);
		if (!ctt_groups[i].htable) {
			perror("hash table initialization");
			exit(1);
		}
		ctt_groups[i].parray = g_ptr_array_new();
		if (!ctt_groups[i].parray) {
			perror("array initialization");
			exit(1);
		}
	}

	atexit(ct_entry_exit);
}

