/*
 * (C) 2010-2011 Florian Westphal <fw@strlen.de>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 */
#include <fcntl.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <libnetfilter_conntrack/libnetfilter_conntrack.h>

#include "ct_entry.h"
#include "ct_opts.h"
#include "parser_common.h"

static struct nfct_handle *nfct_handle;
static struct nfct_handle *nfct_query_h;

static bool mask_match(uint32_t val)
{
	uint32_t mask = cttop_options->ct_event_filter.ctmask;
	return (val & mask) == cttop_options->ct_event_filter.ctmark;
}

static bool event_match_filter(const struct nf_conntrack *ct)
{
	uint32_t mark;

	if (cttop_options->ct_event_filter.family &&
            cttop_options->ct_event_filter.family
		!= nfct_get_attr_u8(ct, ATTR_L3PROTO))
		return false;

	mark = nfct_get_attr_u32(ct, ATTR_MARK);
	if (!mask_match(mark))
		return false;

	/* XXX: match port, protocol, etc. */
	return true;
}

static void nfct2stats(struct nf_conntrack *ct, struct cttop_stats *stats)
{
	struct nfct_attr_grp_ctrs ctrs;
	int ret = nfct_get_attr_grp(ct, ATTR_GRP_ORIG_COUNTERS, &ctrs);
	if (ret == 0) {
		stats->out.bytes = ctrs.bytes;
		stats->out.packets = ctrs.packets;
	} else {
		memset(&stats->out, 0, sizeof(stats->out));
	}
	ret = nfct_get_attr_grp(ct, ATTR_GRP_REPL_COUNTERS, &ctrs);
	if (ret == 0) {
		stats->in.bytes = ctrs.bytes;
		stats->in.packets = ctrs.packets;
	} else {
		memset(&stats->in, 0, sizeof(stats->in));
	}
}

static int event_cb(enum nf_conntrack_msg_type type,
		    struct nf_conntrack *ct,
		    void *data)
{
	static int warn_once;
	struct cttop_stats stats;

	if (!event_match_filter(ct))
		return NFCT_CB_CONTINUE;

	switch (type) {
	case NFCT_T_NEW:
		ct_entry_new(ct);
		break;
	case NFCT_T_DESTROY:
		ct_entry_del(ct);
		break;
	case NFCT_T_UPDATE:
		nfct2stats(ct, &stats);
		ct_entry_update(ct, &stats);
	break;
	default:
		if (warn_once)
			break;
		warn_once = 1;
		fprintf(stderr, "%s: received unknown ct event type %d\n",
						__func__, (int) type);
	}
	(void) data;
	return NFCT_CB_CONTINUE;
}

static int query_cb(enum nf_conntrack_msg_type type,
		    struct nf_conntrack *ct,
		    void *data)
{
	static int warn_once;
	struct cttop_stats stats;

	switch (type) {
	case NFCT_T_UPDATE:
		nfct2stats(ct, &stats);
		ct_entry_update(ct, &stats);
	break;
	case NFCT_T_ERROR:
		fprintf(stderr, "%s: received NFCT_T_ERROR\n",
						__func__);
	break;
	default:
		if (warn_once)
			break;
		warn_once = 1;
		fprintf(stderr, "%s: received unknown ct event type %d\n",
						__func__, (int) type);
	}
	(void) data;
	return NFCT_CB_CONTINUE;
}


bool ctevent_query(struct nf_conntrack *ct)
{
	int ret;

	ret = nfct_query(nfct_query_h, NFCT_Q_GET, ct);
	if (ret == -1) {
		if (errno == EINTR || errno == ENOBUFS || errno == EAGAIN)
			return false;

		if (errno == ENOENT) {
			/*
			 * This happens e.g. if we missed a DELETE message,
			 * or if the delete message has been sent, but is still
			 * unprocessed.
			 */
			ct_entry_del(ct);
			return true;
		}
		perror("nfct_query");
		exit(1);
	}
	return true;
}

void ctevent_dispatch(void)
{
	int ret;

	ret = nfct_catch(nfct_handle);
	if (ret == -1) {
		if (errno == EINTR || errno == EAGAIN || errno == ENOBUFS)
			return;
		perror("nfct_catch");
		exit(1);
	}
}

static void ctevent_exit(void)
{
	nfct_close(nfct_handle);
}

static struct nfct_handle *xnfct_open(uint8_t subsys, unsigned int subscribe)
{
	struct nfct_handle *h;

	h = nfct_open(subsys, subscribe);
	if (!h) {
		perror("nfct_open");
		exit(1);
	}
	return h;
}

static bool fetch_existing_ctrack_family(int family)
{
	int ret;

	/* need to use nfct_handle, so event_cb can filter marks */
        ret = nfct_query(nfct_handle, NFCT_Q_DUMP, &family);
	if (ret != 0)
		perror("nfct_query");
	return ret == 0;
}

static void fetch_existing_ctracks(void)
{
        uint32_t family = cttop_options->ct_event_filter.family;
	if (family == 0)
		family = AF_INET;
	if (!fetch_existing_ctrack_family(family))
		return;
	family = cttop_options->ct_event_filter.family;
	if (family)
		return;

	family = AF_INET6;
	fetch_existing_ctrack_family(family);
}

static void nonblock(int fd)
{
	int flags = fcntl(fd, F_GETFL);
	if (flags == -1) {
		perror("fcntl F_GETFL");
		exit(1);
	}
	if (fcntl(fd, F_SETFL, flags|O_NONBLOCK)) {
		perror("fcntl O_NONBLOCK");
		exit(1);
	}
}

int ctevent_init(void)
{
	nfct_handle = xnfct_open(CONNTRACK, NFCT_ALL_CT_GROUPS);
	nfct_callback_register(nfct_handle, NFCT_T_ALL, event_cb, NULL);

	nfnl_rcvbufsiz(nfct_nfnlh(nfct_handle), 2 * 1024  * 1024);

	fetch_existing_ctracks();

	nfct_query_h = xnfct_open(CONNTRACK, 0);

	nfct_callback_register(nfct_query_h, NFCT_T_ALL, query_cb, NULL);

	bsf_filter_create(nfct_fd(nfct_handle));
	bsf_filter_create(nfct_fd(nfct_query_h));
	nonblock(nfct_fd(nfct_query_h));
	nonblock(nfct_fd(nfct_handle));

	free(cttop_options->ct_event_filter.bsf_str);

	atexit(ctevent_exit);
	return nfct_fd(nfct_handle);
}

