/*
 * (C) 2010-2011 Florian Westphal <fw@strlen.de>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 */
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/epoll.h>
#include <unistd.h>

#include "cttop.h"

#ifdef HAVE_LIBCAP_NG
# include <sys/types.h>
# include <pwd.h>
# include <cap-ng.h>
#endif

#include "dns_cache.h"
#include "ct_entry.h"
#include "ct_events.h"
#include "ct_opts.h"
#include "ct_signal.h"
#include "ct_timer.h"

#include "ui.h"

static int epollfd;

#define MAX_EVENTS 2
#define PROC_SYS_NET_NETFILTER_ACCT	"/proc/sys/net/netfilter/nf_conntrack_acct"

static inline void drop_privileges(void)
{
#ifdef HAVE_LIBCAP_NG
	static const char username[]="nobody";
	uid_t uid = getuid();
	gid_t gid = getgid();

	if (uid == 0) {
		struct passwd *p = getpwnam(username);
		if (!p) {
			fprintf(stderr, "failed to get uid of %s: %s\n",
					username, strerror(errno));
			exit(1);
		}
		uid = p->pw_uid;
		gid = p->pw_gid;
		if (!uid) {
			fprintf(stderr, "uid of %s must not be 0\n", username);
			exit(1);
		}
		endpwent();
	}

	capng_clear(CAPNG_SELECT_BOTH);
	capng_update(CAPNG_ADD, CAPNG_EFFECTIVE|CAPNG_PERMITTED, CAP_NET_ADMIN);
	if (capng_change_id((int) uid, (int) gid,
			CAPNG_DROP_SUPP_GRP | CAPNG_CLEAR_BOUNDING)) {
		perror("Cannot drop privileges");
		exit(1);
	}
#endif
}

static void dispatch(const struct epoll_event *ev)
{
	void(*function)(void);

	function = ev->data.ptr;
	function();
}


static int loop(void)
{
	struct epoll_event events[MAX_EVENTS];
	int nfds;
	int waittime = -1;

	for (;;) {
		nfds = epoll_wait(epollfd, events, ARRAY_SIZE(events), waittime);
		if (nfds < 0 && errno != EINTR) {
			perror("epoll_wait");
			exit(1);
		}
		while (nfds > 0)
			dispatch(&events[--nfds]);
		if (ct_entry_poll_counters_resume())
			waittime = -1;
		else
			waittime = 0;
	}
	return 0;
}

static void wantread(int fd, void (*funct)(void))
{
	struct epoll_event ev = { .events = EPOLLIN };

	ev.data.ptr = funct;

	if (epoll_ctl(epollfd, EPOLL_CTL_ADD, fd, &ev) == -1) {
		perror("epoll_ctl");
		exit(EXIT_FAILURE);
	}
}

static void init_rng(void)
{
	FILE *fp = fopen("/dev/urandom", "rb");
	unsigned int seed = time(NULL);

	if (fp) {
		if (fread(&seed, sizeof(seed), 1, fp) != 1)
			perror("fread");
		fclose(fp);
	}
	srandom(seed);
}

static bool check_nf_acct(void)
{
	FILE *fp = fopen(PROC_SYS_NET_NETFILTER_ACCT, "r");
	int onoff;
	bool ret;

	if (!fp) {
		fprintf(stderr, "%s: %s\n", PROC_SYS_NET_NETFILTER_ACCT, strerror(errno));
		return false;
	}

	ret = false;
	if (1 == fscanf(fp, "%d", &onoff) && onoff == 1)
		ret = true;
	fclose(fp);
	return ret;
}

int main(int argc, char *argv[])
{
	int fd;

	drop_privileges();

	if (!check_nf_acct()) {
		fprintf(stderr, "Warning: %s must be set to 1 for conntrack counter support\n",
				PROC_SYS_NET_NETFILTER_ACCT);
		sleep(5);
	}

	cttop_parse_options(argc, argv);

	epollfd = epoll_create(10);
	if (epollfd < 0) {
		perror("epoll_create");
		exit(1);
	}

	init_rng();

	ct_entry_init();
	fd = ctevent_init();
	wantread(fd, ctevent_dispatch);
	fd = ctsignal_init();
	wantread(fd, ctsignal_dispatch);

	ui_init();

	if (!cttop_options->batch_mode)
		wantread(0, ui_dispatch);

	fd = cttimer_init();
	wantread(fd, cttimer_dispatch);

	fd = dnscache_init();
	if (fd >= 0)
		wantread(fd, dnscache_dispatch);
	loop();
	return 0;
}
